import os
from typing import Callable, List, Optional

import torch

from torch_geometric.data import (
    Data,
    InMemoryDataset,
    download_url,
    extract_zip,
)


class custom_dataset(InMemoryDataset):
    url = "temp"
    def __init__(
        self,
        name,
        url,
        root: str,
        train: bool = -1,
        transform: Optional[Callable] = None,
        pre_transform: Optional[Callable] = None,
        pre_filter: Optional[Callable] = None,
    ):
        self.name = name
        self.url = url
        self.train = train
        super().__init__(root, transform, pre_transform, pre_filter)

        if train == True:
            path = self.processed_paths[0]
        elif train == False:
            path = self.processed_paths[1]
        elif train == -1:
            path = self.processed_paths[2]

        self.path = path
        self.data, self.slices = torch.load(path)

    @property
    def raw_file_names(self) -> str:
        return self.name + '.pt'

    @property
    def processed_file_names(self) -> List[str]:
        return ['train_data.pt', 'test_data.pt','all_data.pt']

    def download(self):
        path = download_url(self.url, self.raw_dir)
        extract_zip(path, self.raw_dir)
        os.unlink(path)

    def process(self):
        inputs = torch.load(self.raw_paths[0])
        inputs.process()
        self.data = inputs




import torch
from torch_geometric.data import Dataset, Data
import numpy as np 
import os
from torch_geometric.data import (
    Data,
    InMemoryDataset,
    download_url,
    extract_zip,
)

CLASS_NAME = "DISEASE"

class CytokinesDataSet(InMemoryDataset):
    def __init__(self, root, filename, graphName, test=False, transform=None, pre_transform=None, patients = None, adjacency = None, nodeNames = None, divisions = None):
        """
        root = Where the dataset should be stored. This folder is split
        into raw_dir (downloaded dataset) and processed_dir (processed data). 
        """

        self.test = test
        self.filename = filename
        self.patients = patients
        self.adjacency = adjacency
        self.graphName = graphName
        self.nodeNames = nodeNames
        self.divisions = divisions # 
        self.process()


        super(CytokinesDataSet, self).__init__(root, transform, pre_transform)
        
    @property
    def raw_file_names(self):
        """ If this file exists in raw_dir, the download is not triggered.
            (The download func. is not implemented here)  
        """
        return self.filename

    @property
    def processed_file_names(self):
        """ If these files are found in raw_dir, processing is skipped. Here, always process

        fileNameArray = []
        if self.test:
            for i in range(len(self.patients)):
                fileNameArray.append(graphName[:-9] + "_data_test_" + str(i) + ".pt")
        else:
            for i in range(len(self.patients)):
                fileNameArray.append(graphName[:-9] + "_data_test_" + str(i) + ".pt")"""
        
        fileNameArray = ["dummy.csv"]
        
        return fileNameArray

    def download(self):
        pass

    def process(self):
        if '_graph.csv' in self.graphName:
            name = self.graphName[:self.graphName.index("_graph.csv")]
        else:
            name = self.graphName

        self.new_dir = "datasets/" + name + "/processed/"# "datasets\\" + name + "\\processed"   # new processed dir
        train_data = Data()
        train_slices = dict()
        test_data = Data()
        test_slices = dict()
        all_data = Data()
        all_slices = dict()
        train_tuple = (train_data, train_slices)
        test_tuple = (test_data, test_slices)
        all_tuple = (all_data, all_slices)

        num_patients = len(self.patients)
        num_training = int(0.8*num_patients) # when index <= num_training, the patient goes into training data. Else, testing
        num_testing = num_patients - num_training
        num_nodes = len(self._get_node_features(self.patients[0]))
        num_edges = len(self.adjacency)
        
        
        # prepare the slices
        train_y_slice = []
        train_x_slice = []
        train_edge_index_slice = []

        test_y_slice = []
        test_x_slice = []
        test_edge_index_slice = []

        all_y_slice = []
        all_x_slice = []
        all_edge_index_slice = []

        for i in range(num_training + 1):
            train_y_slice.append(i)
            train_x_slice.append(num_nodes*i)
            train_edge_index_slice.append(num_edges * i)

        for i in range(num_testing + 1):
            test_y_slice.append(i)
            test_x_slice.append(num_nodes*i)
            test_edge_index_slice.append(num_edges * i)

        for i in range(num_patients + 1):
            all_y_slice.append(i)
            all_x_slice.append(num_nodes*i)
            all_edge_index_slice.append(num_edges * i)
        
        train_slices['y'] = torch.tensor(train_y_slice)
        train_slices['x'] = torch.tensor(train_x_slice)
        train_slices['edge_index'] = torch.tensor(train_edge_index_slice)

        test_slices['y'] = torch.tensor(test_y_slice)
        test_slices['x'] = torch.tensor(test_x_slice)
        test_slices['edge_index'] = torch.tensor(test_edge_index_slice)


        all_slices['y'] = torch.tensor(all_y_slice)
        all_slices['x'] = torch.tensor(all_x_slice)
        all_slices['edge_index'] = torch.tensor(all_edge_index_slice)

        
        node_vector_len = len(self._get_node_features(self.patients[0]).numpy()[0]) # the length of a node vector
        # preparing data
        train_x_tensor = torch.empty((0, node_vector_len), dtype=torch.float32)
        train_edge_index_tensor = torch.empty((2, 0), dtype=torch.int32)
        train_y_list = [] # It's just easier to make a list than figure out how to do 1d concats

        test_x_tensor = torch.empty((0, node_vector_len), dtype=torch.float32)
        test_edge_index_tensor = torch.empty((2, 0), dtype=torch.int32)
        test_y_list = [] # It's just easier to make a list than figure out how to do 1d concats


        all_x_tensor = torch.empty((0, node_vector_len), dtype=torch.float32)
        all_edge_index_tensor = torch.empty((2, 0), dtype=torch.int32)
        all_y_list = [] # It's just easier to make a list than figure out how to do 1d concats


        # runs once for each patient
        index = 0
        for patient in self.patients:
            # Get node features of a single patient
            node_feats = self._get_node_features(patient)
            # Get edge features of a single patient
            edge_feats = self._get_edge_features()
            # Get adjacency info of a single patient
            edge_index = self._get_adjacency_info()
            # Get labels info of a single patient
            label = self._get_labels(patient[CLASS_NAME])

            if index <= num_training:
                train_x_tensor = torch.cat((train_x_tensor,node_feats), 0)
                train_edge_index_tensor = torch.cat((train_edge_index_tensor,edge_index), 1)
                train_y_list.append(int(patient[CLASS_NAME]))
            else:
                test_x_tensor = torch.cat((test_x_tensor,node_feats), 0)
                test_edge_index_tensor = torch.cat((test_edge_index_tensor,edge_index), 1)
                test_y_list.append(int(patient[CLASS_NAME]))
            
            all_x_tensor = torch.cat((all_x_tensor,node_feats), 0)
            all_edge_index_tensor = torch.cat((all_edge_index_tensor,edge_index), 1)
            all_y_list.append(int(patient[CLASS_NAME]))

            index += 1

        # turn y_lists into tensors
        train_y_tensor = torch.tensor(train_y_list)
        test_y_tensor = torch.tensor(test_y_list)
        all_y_tensor = torch.tensor(all_y_list)

        train_data.x = train_x_tensor
        train_data.y = train_y_tensor
        train_data.edge_index = train_edge_index_tensor


        test_data.x = test_x_tensor
        test_data.y = test_y_tensor
        test_data.edge_index = test_edge_index_tensor


        all_data.x = all_x_tensor
        all_data.y = all_y_tensor
        all_data.edge_index = all_edge_index_tensor


        torch.save(self.divisions, os.path.join(self.new_dir, 'divisions.pt')) # states which genes belong to which cytokine
        torch.save(self.nodeNames, os.path.join(self.new_dir, 'nodeNames.pt'))
        torch.save(train_tuple, os.path.join(self.new_dir, 'train_data.pt'))
        torch.save(test_tuple, os.path.join(self.new_dir, 'test_data.pt'))
        torch.save(all_tuple, os.path.join(self.new_dir, 'all_data.pt'))


        # last step, just save everything in the right .pt files
        """
                return ['train_data.pt', 'test_data.pt']

                        torch.save(data, 
                    os.path.join(self.processed_dir, 
                                 f'{name}_data_{index}.pt'))
        """

        

    def _get_node_features(self, patient):
        """ 
        This will return a matrix / 2d array of the shape
        [Number of Nodes, Node Feature size]
        """
        all_node_feats = np.asarray(patient['data'])

        return torch.tensor(all_node_feats, dtype=torch.float)

    # todo: REturn list of 1 each time
    def _get_edge_features(self):
        """ 
        This will return a matrix / 2d array of the shape
        [Number of edges, Edge Feature size]
        """
        all_edge_feats = []
        for edge in self.adjacency:
            all_edge_feats.append(1)

        all_edge_feats = np.asarray(all_edge_feats)
        return torch.tensor(all_edge_feats, dtype=torch.float)

    # todo, same adjacency
    def _get_adjacency_info(self):
        """
        We could also use rdmolops.GetAdjacencyMatrix(mol)
        but we want to be sure that the order of the indices
        matches the order of the edge features
        """


        edge_indices = []
        for edge in self.adjacency:
            edge_indices.append(edge[0])

        for edge in self.adjacency:
            edge_indices.append(edge[1])

        edge_indices = torch.tensor(edge_indices)
        edge_indices = edge_indices.t().to(torch.long).view(2, -1)
        return edge_indices

    
    def _get_labels(self, label):
        return torch.tensor(float(label), dtype=torch.int64)

    def len(self):
        return len(self.patients)
    
    def get(self, idx: int):
        pass


import logging
import os

import torch
from torch_geometric import seed_everything
from deepsnap.dataset import GraphDataset
from graphgym.cmd_args import parse_args
from graphgym.config import cfg, dump_cfg, load_cfg, set_run_dir, set_out_dir
from graphgym.loader import create_dataset, create_loader
from graphgym.logger import create_logger, setup_printing
from graphgym.model_builder import create_model
from graphgym.optimizer import create_optimizer, create_scheduler
from graphgym.register import train_dict
from graphgym.train import train
from graphgym.utils.agg_runs import agg_runs
from graphgym.utils.comp_budget import params_count
from graphgym.utils.device import auto_select_device
from graphgym.models.gnn import GNNStackStage
from iCYPRESS.CytokinesDataSet import CytokinesDataSet
from iCYPRESS.Visualization import Visualize
from graphgym.models.layer import GeneralMultiLayer, Linear, GeneralConv
from graphgym.models.gnn import GNNStackStage

class Cypress:
    def __init__(self):
        pass
    
    def train(self):
        # Load cmd line args
        args = parse_args()
        # Load config file
        load_cfg(cfg, args)
        set_out_dir(cfg.out_dir, args.cfg_file)
        # Set Pytorch environment
        torch.set_num_threads(cfg.num_threads)
        dump_cfg(cfg)
        # Repeat for different random seeds
        for i in range(args.repeat):
            set_run_dir(cfg.out_dir)
            setup_printing()
            # Set configurations for each run
            cfg.seed = cfg.seed + 1
            seed_everything(cfg.seed)
            auto_select_device()
            # Set machine learning pipeline
            datasets = create_dataset()
            loaders = create_loader(datasets)
            loggers = create_logger()
            model = create_model()





            # Add edge_weights attribute to the datasets so that they can be accessed in batches
            num_edges = len(datasets[0][0].edge_index[0])
            edge_weights = torch.nn.Parameter(torch.ones(num_edges))
            for loader in loaders:
                for dataset in loader.dataset:
                    dataset.edge_weights = edge_weights


            #add edge weights to the set of parameters
            newParam = list()
            for param in model.parameters():
                newParam.append(param)
            
            newParam.append(edge_weights)

            optimizer = create_optimizer(newParam)
            scheduler = create_scheduler(optimizer)
            # Print model info
            logging.info(model)
            logging.info(cfg)
            cfg.params = params_count(model)
            logging.info('Num parameters: %s', cfg.params)


            # Start training
            if cfg.train.mode == 'standard':
                train(loggers, loaders, model, optimizer, scheduler)
            else:
                train_dict[cfg.train.mode](loggers, loaders, model, optimizer,
                                        scheduler)
            # Aggregate results from different seeds
            agg_runs(cfg.out_dir, cfg.metric_best)
            # When being launched in batch mode, mark a yaml as done
            if args.mark_done:
                os.rename(args.cfg_file, f'{args.cfg_file}_done')


import logging
import os

import torch
from torch_geometric import seed_everything
from deepsnap.dataset import GraphDataset
from graphgym.custom_dataset import custom_dataset
from graphgym.cmd_args import parse_args
from graphgym.config import cfg, dump_cfg, load_cfg, set_run_dir, set_out_dir
from graphgym.loader import create_dataset, create_local_dataset, create_loader
from graphgym.logger import create_logger, setup_printing
from graphgym.model_builder import create_model
from graphgym.optimizer import create_optimizer, create_scheduler
from graphgym.register import train_dict
from graphgym.train import train
from graphgym.utils.agg_runs import agg_runs
from graphgym.utils.comp_budget import params_count
from graphgym.utils.device import auto_select_device
from graphgym.models.gnn import GNNStackStage
from iCYPRESS.CytokinesDataSet import CytokinesDataSet
from iCYPRESS.Visualization import Visualize
from graphgym.models.layer import GeneralMultiLayer, Linear, GeneralConv
from graphgym.models.gnn import GNNStackStage
import numpy as np


class Cypress:
    def __init__(self, patients = None, eset = None, blood_only=True, active_cyto_list = ['CCL1'], batch_size = 80, 
                 eval_period = 20, layers_pre_mp = 2, layers_mp = 6, layers_post_mp = 2, 
                 dim_inner = 137, max_epoch = 400):
        
        if eset is None:
            self.makeExistingConfigFile()
            self.custom = False
            return
        
        self.custom = True
        # get current config file
        current_dir = os.path.dirname(os.path.abspath(__file__))
        curr_cfg_file = os.path.join(current_dir, "configs", "example_custom.yaml")
        self.write_lines_to_file(curr_cfg_file, os.path.join( "configs", "example_custom.yaml"))
        self.current_dir = os.path.dirname(os.path.abspath(__file__))

        self.eset = eset

        patient_dict, patient_list = self.process_patients(patients) # a dict that matches a patient name to their classification

        
        cyto_list,cyto_adjacency_dict,cyto_tissue_dict  = self.process_graphs(blood_only) # list of cytokines, maps a cytokine's name to their adjacency matrix, maps a cytokine's name to the tissues they need


        tissue_gene_dict, gene_set = self.process_tissues(blood_only) # dict that matches tissues to the genes associated with them, a set of all genes we have


        gene_to_patient, active_tissue_gene_dict = self.process_eset(eset, gene_set, patient_dict, tissue_gene_dict) # 2 layer deep dict. First layer maps gene name to a dict. Second layer matches patient code to gene expresion data of the given gene.
        

        # now convert these to vectors like needed
        # probably best to only do this for ONE cytokine at a time? Yeah, accept a gene, then move.
        # also, make a function that prints out everu available cytokine.
        for cyto in active_cyto_list:
            if cyto not in cyto_tissue_dict.keys():
                raise(ValueError("{} is not a valid cytokine.".format(cyto)))
            self.create_cyto_database(cyto, eset, cyto_tissue_dict, active_tissue_gene_dict, patient_list, 
                                     patient_dict, gene_to_patient, cyto_adjacency_dict)
           
            name = cyto + "_" + eset[:eset.index(".")]  
            self.makeConfigFile(name, batch_size, eval_period, layers_pre_mp, layers_mp, 
                       layers_post_mp, dim_inner, max_epoch)
        
            self.active_cyto_list = active_cyto_list

    def makeExistingConfigFile(self):
        if (not os.path.exists(os.path.abspath("configs"))):
            os.makedirs(os.path.abspath("configs"))
        
        # get current config file
        current_dir = os.path.dirname(os.path.abspath(__file__))
        curr_cfg_file = os.path.join(current_dir, "configs", "example_custom.yaml")
        self.write_lines_to_file(curr_cfg_file, os.path.join("configs", "example_custom.yaml"))

    def create_cyto_database(self, cyto, eset, cyto_tissue_dict, active_tissue_gene_dict, patient_list, 
                             patient_dict, gene_to_patient, cyto_adjacency_dict):

        # creates graphname
        graphName = cyto + "_" + eset[:eset.index(".")]

        #create patientArray
        patientArray = []

        # count the number of active genes in each tissue.
        tissues = cyto_tissue_dict[cyto]

        gene_count = []
        for tissue in tissues:
            count = len(active_tissue_gene_dict[tissue])
            
            gene_count.append(count)
        
        total_genes = sum(gene_count)


        for patient in patient_list:
            patient_data = {}

            patient_data["DISEASE"] = str(patient_dict[patient])


            data = []
            # create the information that goes into each node
            for i, tissue in enumerate(tissues):
                tissue_data = [0]*total_genes
                start = sum(gene_count[:i])

                tissue_genes = active_tissue_gene_dict[tissue]
                
                offset = 0
                for gene in tissue_genes:                        
                    tissue_data[start + offset] = gene_to_patient[gene][patient] / 20
                    offset +=  1
                
                data.append(tissue_data)

            patient_data["data"] = data
            patientArray.append(patient_data)

                
        nodeList = cyto_tissue_dict[cyto]
        graphAdjacency = cyto_adjacency_dict[cyto]

        nodeIntMap = {}
        i = 0

        for node in nodeList:
            nodeIntMap[node] = i
            i += 1

        intAdjacency = []
        # turn the adjacency names into int
        for edge in graphAdjacency:
            newEdge = [nodeIntMap[edge[0]], nodeIntMap[edge[1]]]
            intAdjacency.append(newEdge)

        try:
            os.mkdir(os.path.join("datasets"))
        except OSError:
            pass
    
        try:
            os.mkdir(os.path.join("datasets", graphName))
        except OSError:
            pass

        try:
            os.mkdir(os.path.join("datasets", graphName ,"raw"))
        except OSError:
            pass

        try:
            os.mkdir(os.path.join("datasets", graphName, "processed"))
        except OSError:
            pass
            
        full_dataset = CytokinesDataSet(root="data/", graphName=graphName, filename="full.csv", 
                                        test=True, patients=patientArray, adjacency=intAdjacency, 
                                        nodeNames = nodeIntMap, divisions = gene_count)
        

        torch.save(full_dataset, (os.path.join("datasets", graphName, "raw",  graphName + ".pt")))



    def process_eset(self, eset, gene_set, patient_dict, tissue_gene_dict):
        eset_file = open(eset, 'r')

        eset_lines = eset_file.read().splitlines()

        
        # read the first line, and see if it matches with the patient file provided
        patients = eset_lines[0].replace("\"", "").split(",")[2:]
        
        patient_set = set(patient_dict.keys())

        for patient in patients:
            try:
                patient_set.remove(patient)
            except(KeyError):
                raise(ValueError("{} is not found in the patients file.".format(patient)))


        if (len(patient_set) != 0):
            raise(ValueError("The eset file does not contain {}".format(patient_set)))


        gene_to_patient = dict() # maps the name of a gene to a dict of patients
        for line_num in range(1, len(eset_lines)):
            line = eset_lines[line_num].replace("\"", "")
            parts = line.split(",")
            new_gene = parts[1]


            if (new_gene not in gene_set):
                continue
            # get all the gene expression numbers, and then normalize them
            gene_nums = parts[2:]
            gene_nums = [float(gene_num) for gene_num in gene_nums]
            gene_nums = self.normalize_vector(gene_nums)


            
            patient_gene_data_dict = dict() # maps the patients code to their gene expression data of this one specific gene
            for index, patient in enumerate(patients):
                patient_gene_data_dict[patient] = gene_nums[index]

            gene_to_patient[new_gene] = patient_gene_data_dict
            
        # make a new tissue_gene_dict

        active_tissue_gene_dict = dict()

        for tissue in tissue_gene_dict.keys():
                gene_array = []

                original_genes = tissue_gene_dict[tissue]

                for gene in original_genes:
                    if gene in gene_to_patient.keys():
                        gene_array.append(gene)
                
                active_tissue_gene_dict[tissue] = gene_array
        return (gene_to_patient, active_tissue_gene_dict)

    def normalize_vector(self, vector):
        min_val = np.min(vector)
        max_val = np.max(vector)
        normalized_vector = (vector - min_val) / (max_val - min_val)
        return normalized_vector    
    
    def process_patients(self, patients):
        patient_file = open(patients, 'r')
        patient_dict = dict()
        patient_list = []

        for line in patient_file.read().splitlines():
            parts = line.split(",")
            patient_dict[parts[0]] = int(parts[1])
            patient_list.append(parts[0])
        
        return patient_dict, patient_list
        

    def process_graphs(self, blood_only):
        if(blood_only):
            graph_folder_path = os.path.join(self.current_dir,"Modified Graphs")
        else:
            graph_folder_path = os.path.join(self.current_dir,"Graphs")
        
        cyto_list = []
        cyto_adjacency_dict = dict() # maps a cytokine's name to their adjacency matrix
        cyto_tissue_dict = dict() # maps a cytokine's name to the tissues they need

        for filename in os.listdir(graph_folder_path):
            cyto_name = filename[:-10]
            cyto_list.append(cyto_name) # drop the _graph.csv
            graph_file_path = os.path.join(graph_folder_path, filename)
            graphAdjacency = []
            tissue_set = set()

            f = open(graph_file_path, 'r')
            graphLines = f.read().splitlines()
            
            for line in graphLines:
                parts = line.upper().split(",") # remove newline, capitalize, and remove spaces
                graphAdjacency.append(parts)
                newParts = [parts[1], parts[0]]
                tissue_set.update(newParts)

                graphAdjacency.append(newParts)
            

            # put the tissues into a list, and then sort them
            tissue_list = []
            for tissue in tissue_set:
                tissue_list.append(tissue)
            
            tissue_list.sort()

            cyto_adjacency_dict[cyto_name] = graphAdjacency
            cyto_tissue_dict[cyto_name] = tissue_list


        return cyto_list,cyto_adjacency_dict,cyto_tissue_dict

    def process_tissues(self, blood_only):
        tissue_gene_dict = dict() # maps tissues to the genes associated with them

        gene_set = set()

        tissue_file = open(os.path.join(self.current_dir, "GenesToTissues.csv"))

        tissue_lines = tissue_file.read().splitlines()

        for i in range(0, len(tissue_lines), 2):
            tissue_line = tissue_lines[i]
            
            tissue_line_arr = tissue_line.split(",")

            if (blood_only and (tissue_line_arr[1] == "N")):
                continue

            tissue = tissue_line_arr[0]
            genes_array = tissue_lines[i + 1].split(',')

            tissue_gene_dict[tissue] = genes_array

            gene_set.update(genes_array)

        return tissue_gene_dict, gene_set
    

    def makeConfigFile(self, name, batch_size, eval_period, layers_pre_mp, layers_mp, 
                       layers_post_mp, dim_inner, max_epoch):
        if (not os.path.exists(os.path.abspath("configs"))):
            os.makedirs(os.path.abspath("configs"))

        # using with statement
        with open(os.path.join('configs', name + ".yaml"), 'w') as file:
            file.write('out_dir: results\n')
            file.write('dataset:\n')
            file.write(' format: PyG\n')
            file.write(' name: ' + name + '\n')
            file.write(' task: graph\n')
            file.write(' task_type: classification\n')
            file.write(' transductive: True\n')
            file.write(' split: [0.8, 0.2]\n')
            file.write(' augment_feature: []\n')
            file.write(' augment_feature_dims: [0]\n')
            file.write(' augment_feature_repr: position\n')
            file.write(' augment_label: \'\'\n')
            file.write(' augment_label_dims: 0\n')
            file.write(' transform: none\n')
            file.write('train:\n')
            file.write(' batch_size: ' + str(batch_size) + '\n' )
            file.write(' eval_period: ' + str(eval_period) + '\n')
            file.write(' ckpt_period: 100\n')
            file.write('model:\n')
            file.write(' type: gnn\n')
            file.write(' loss_fun: cross_entropy\n')
            file.write(' edge_decoding: dot\n')
            file.write(' graph_pooling: add\n')
            file.write('gnn:\n')
            file.write(' layers_pre_mp: ' + str(layers_pre_mp) + '\n')
            file.write(' layers_mp: ' + str(layers_mp) + '\n')
            file.write(' layers_post_mp: ' + str(layers_post_mp) + '\n')
            file.write(' dim_inner: ' + str(dim_inner) + '\n')
            file.write(' layer_type: generalconv\n')
            file.write(' stage_type: skipsum\n')
            file.write(' batchnorm: True\n')
            file.write(' act: prelu\n')
            file.write(' dropout: 0.0\n')
            file.write(' agg: add\n')
            file.write(' normalize_adj: False\n')
            file.write('optim:\n')
            file.write(' optimizer: adam\n')
            file.write(' base_lr: 0.01\n')
            file.write(' max_epoch: ' +  str(max_epoch) + '\n')


    def write_lines_to_file(self, input_file, output_file_name):
        try:
            with open(input_file, 'r') as infile:
                # Read all lines from the input file
                lines = infile.readlines()

            # Remove any leading/trailing whitespaces from the output file name
            output_file_name = output_file_name.strip()

            # Write the lines to the output file with the provided name (overwriting if it exists)
            with open(output_file_name, 'w') as outfile:
                for line in lines:
                    outfile.write(line)

        except FileNotFoundError:
            print(f"Error: Input file '{input_file}' not found.")
        except Exception as e:
            print(f"Error: An unexpected error occurred - {e}")

    def train(self, cyto = None):


        if (self.custom):
            # check if cytokine has been initalized
            if cyto not in self.active_cyto_list:
                raise(ValueError("{} has not been initalized as a cytokine".format(cyto)))

            name = cyto + "_" + self.eset[:self.eset.index(".")]  

            args = parse_args(name + ".yaml")
        else:
            name = None
            args = parse_args()
            
        # Load config file
        load_cfg(cfg, args)
        
        set_out_dir(cfg.out_dir, args.cfg_file)
        # Set Pytorch environment
        torch.set_num_threads(cfg.num_threads)
        dump_cfg(cfg)
        # Repeat for different random seeds
        for i in range(args.repeat):
            set_run_dir(cfg.out_dir)
            setup_printing()
            # Set configurations for each run
            cfg.seed = cfg.seed + 1
            seed_everything(cfg.seed)
            auto_select_device()
            # Set machine learning pipeline
            # to fix this, copy create_dataset, only with formed graphs already created.

            if (self.custom):
                dataset_raw = custom_dataset(root=os.path.join("dataset",  name), name=name, url="null")
                graphs = GraphDataset.pyg_to_graphs(dataset_raw)
                datasets = create_local_dataset(graphs)
            else:
                datasets = create_dataset()
            loaders = create_loader(datasets)
            loggers = create_logger()
            model = create_model()





            # Add edge_weights attribute to the datasets so that they can be accessed in batches
            num_edges = len(datasets[0][0].edge_index[0])
            edge_weights = torch.nn.Parameter(torch.ones(num_edges))
            for loader in loaders:
                for dataset in loader.dataset:
                    dataset.edge_weights = edge_weights


            #add edge weights to the set of parameters
            newParam = list()
            for param in model.parameters():
                newParam.append(param)
            
            newParam.append(edge_weights)

            optimizer = create_optimizer(newParam)
            scheduler = create_scheduler(optimizer)
            # Print model info
            logging.info(model)
            logging.info(cfg)
            cfg.params = params_count(model)
            logging.info('Num parameters: %s', cfg.params)


            # Start training
            if cfg.train.mode == 'standard':
                train(loggers, loaders, model, optimizer, scheduler)
            else:
                train_dict[cfg.train.mode](loggers, loaders, model, optimizer,
                                        scheduler)
            # Aggregate results from different seeds
            agg_runs(cfg.out_dir, cfg.metric_best)
            # When being launched in batch mode, mark a yaml as done
            if args.mark_done:
                os.rename(args.cfg_file, f'{args.cfg_file}_done')

        

    def get_cyto_list():
        return ['CCL1', 'CCL2', 'CCL3', 'CCL3L1', 'CCL4', 'CCL4L2', 'CCL5', 'CCL7', 
                'CCL8', 'CCL13', 'CCL17', 'CCL19', 'CCL20', 'CCL23', 'CCL24', 'CCL25', 
                'CCL26', 'CCL27', 'CD40LG', 'CD70', 'CKLF', 'CSF1', 'CSF2', 'CXCL2', 'CXCL8', 
                'CXCL9', 'CXCL10', 'CXCL11', 'CXCL12', 'CXCL13', 'CXCL14', 'EGF', 'FASLG', 
                'FLT3LG', 'HGF', 'IFNG', 'IFNL1', 'IL2', 'IL3', 'IL4', 'IL5', 'IL6', 
                'IL7', 'IL11', 'IL12A', 'IL12B', 'IL18', 'IL19', 'IL1B', 'IL1RN', 'IL23A', 
                'IL26', 'LIF', 'OSM', 'PDGFB', 'PF4', 'PPBP', 'SPP1', 'TGFA', 'TGFB1', 
                'TGFB3', 'TNFSF9', 'TNFSF10', 'TNFSF11', 'TNFSF13', 'TNFSF13B', 'TNFSF14', 
                'TNF', 'XCL1', 'XCL2']


import argparse


def parse_args(name = "example_custom.yaml" ) -> argparse.Namespace:
    r"""Parses the command line arguments."""
    parser = argparse.ArgumentParser(description='GraphGym')

    parser.add_argument('--cfg',
                        dest='cfg_file',
                        type=str,
                        default='configs/' + name,
                        help='The configuration file path.')
    parser.add_argument('--repeat',
                        type=int,
                        default=1,
                        help='The number of repeated jobs.')
    parser.add_argument('--mark_done',
                        action='store_true',
                        default=[],
                        help='Mark yaml as done after a job has finished.')
    parser.add_argument('opts',
                        default=None,
                        nargs=argparse.REMAINDER,
                        help='See graphgym/config.py for remaining options.')

    return parser.parse_args()



import torch
import matplotlib.pyplot as plt
import networkx as nx
from torch_geometric.utils import to_networkx
from sklearn.manifold import TSNE

class Visualize:
    def visualize_correlations(name, graph, correlationMatrix):
        nodeNameLocation = "datasets\\" + name + "\\processed\\nodeNames.pt"
        nodeNames = torch.load(nodeNameLocation)
        graphCopy = nx.create_empty_copy(graph)
        cutoff = 0.9

        inv_map = {v: k for k, v in nodeNames.items()}
        num_nodes = len(graphCopy.nodes)


        for row in range(num_nodes):
            for col in range(row + 1, num_nodes):
                if(correlationMatrix[row][col] >= cutoff):
                    graphCopy.add_edge(row, col)


        graphCopy = nx.relabel_nodes(graphCopy, inv_map)


        pos = nx.shell_layout(graphCopy)
        nx.draw_networkx_nodes(graphCopy, pos, cmap=plt.get_cmap('jet'))
        nx.draw_networkx_labels(graphCopy, pos, font_color = "red")
        nx.draw_networkx_edges(graphCopy, pos, edgelist=graphCopy.edges)
        plt.show()
        return

    def visualize_TSNE(matrix, classification):



        # Apply t-SNE to the matrix
        tsne = TSNE(n_components=2)
        embedded_matrix = tsne.fit_transform(matrix)

        # Plot the points based on the classification array
        plt.scatter(embedded_matrix[:, 0], embedded_matrix[:, 1], c=classification)
        plt.colorbar()
        plt.show()
        return

    def visualize_graph(colorWeights, graph, name, edge_weights):
        nodeNameLocation = "datasets\\" + name + "\\processed\\nodeNames.pt"
        divisionsLocation = "datasets\\" + name + "\\processed\\divisions.pt"
        nodeNames = torch.load(nodeNameLocation)
        divisions = torch.load(divisionsLocation)

        largest_edge = max(edge_weights).item()
        smallest_edge = min(edge_weights).item()

        diff_edge = largest_edge - smallest_edge

        edge_colours = []

        for edge in edge_weights:
            color = [0]*3
            if((edge.item() - smallest_edge) / diff_edge > 0.9):
                color[1] = 1.0
            edge_colours.append(color)




        inv_map = {v: k for k, v in nodeNames.items()}

        graph = nx.relabel_nodes(graph, inv_map)
        graph = graph.to_directed()






        sampleSource = Visualize.nodeColourings(colorWeights, divisions)
        num_edges = len(graph.edges)
        
        values = []

        for i in range(len(graph.nodes)):
            values.append(sampleSource[i])

        pos = nx.shell_layout(graph)
        nx.draw_networkx_nodes(graph, pos, cmap=plt.get_cmap('jet'),  node_color = values, node_size = 1000)
        nx.draw_networkx_labels(graph, pos, font_color = "red")
        nx.draw_networkx_edges(graph, pos, edgelist=graph.edges, edge_color=edge_colours[:num_edges], node_size = 1000, arrows=True, connectionstyle='arc3, rad = 0.1')
        plt.show()
    
    """
    A function that returns how each node is to be colored. 
    colorWeights is the weights of the layer we are looking at.
    division is an array that tells us which weights belong to which node in the graph
    """
    def nodeColourings(colorWeights, division):
        # not sure if I like a simple average eqaution. 

        num_rows = len(colorWeights)
        num_cols = len(colorWeights[0])


        # stores the average value of the weight of a single gene
        averages = [None] * num_cols
        for col in range(num_cols):
            total = 0
            for row in range(num_rows):
                total += abs(colorWeights[row][col].item())
            
            averages[col] = total / num_rows
            
        # we have an array that has the average of every gene. Now we need to average across nodes
        node_avg = []
        start = 0
        for end in division:
            total = 0
            for i in range(start, end): #average out, then print list, then return
                total += averages[i]
            average = total / (end - start)
            node_avg.append(average)
            start = end
        

        largest_avg = 0

        # find the largest average to use as scaling
        for avg in node_avg:
            largest_avg = max(largest_avg, abs(avg))
        

        sampleSource = []
        for i in range(len(node_avg)):
            color = [0]*3
            
            color[2] = node_avg[i] / largest_avg

            sampleSource.append(color)





        

        return sampleSource

